﻿// TouchThumb.cs (c) 2009 by Charles Petzold
using System;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Controls.Primitives;
using System.Windows.Input;
using System.Windows.Media;

namespace Petzold.Controls
{
    public class TouchThumb : Control
    {
        // Public events
        public event DragStartedEventHandler DragStarted;
        public event DragDeltaEventHandler DragDelta;
        public event DragCompletedEventHandler DragCompleted;
        public static event TouchFrameEventHandler UnhandledFrameReported;

        // Static fields (for all ThouchThumb objects)
        static List<TouchThumb> touchThumbs = new List<TouchThumb>();
        static Dictionary<int, Info> infoDictionary = new Dictionary<int, Info>();

        // Instance fields
        bool isMouseCaptured;
        Point mousePosition;

        // Static constructor: Just attach handler for Touch
        static TouchThumb()
        {
            Touch.FrameReported += OnTouchFrameReported;
        }

        // Event Handler for Touch
        static void OnTouchFrameReported(object sender, TouchFrameEventArgs args)
        {
            TouchPoint primaryTouchPoint = args.GetPrimaryTouchPoint(null);
            bool isPrimaryDown = (primaryTouchPoint != null && primaryTouchPoint.Action == TouchAction.Down);

            if (isPrimaryDown)
                args.SuspendMousePromotionUntilTouchUp();

            TouchPointCollection touchPoints = args.GetTouchPoints(null);

            foreach (TouchPoint touchPoint in touchPoints)
            {
                TouchThumb touchThumb = null;
                int deviceId = touchPoint.TouchDevice.Id;

                switch (touchPoint.Action)
                {
                    case TouchAction.Down:
                        touchThumb = GetTouchThumbAncestor(touchPoint.TouchDevice.DirectlyOver);

                        if (touchThumb != null)
                        {
                            Info info = new Info()
                            {
                                TouchThumb = touchThumb,
                                Position = touchPoint.Position
                            };

                            infoDictionary.Add(touchPoint.TouchDevice.Id, info);

                            // Get a position relative to the TouchThumb for this event
                            Point pt = touchThumb.TransformToVisual(null).Transform(new Point());
                            touchThumb.OnDragStarted(touchPoint.Position.X - pt.X, 
                                                     touchPoint.Position.Y - pt.Y);
                        }
                        // Fire UnhandledFrameReported event
                        else if (isPrimaryDown)
                        {
                            if (UnhandledFrameReported != null)
                                UnhandledFrameReported(sender, args);
                        }
                        break;

                    case TouchAction.Move:
                        if (infoDictionary.ContainsKey(deviceId))
                            touchThumb = infoDictionary[deviceId].TouchThumb;

                        if (touchThumb != null)
                        {
                            Point oldPosition = infoDictionary[deviceId].Position;
                            Point newPosition = touchPoint.Position;
                            infoDictionary[deviceId].Position = newPosition;

                            touchThumb.OnDragDelta(newPosition.X - oldPosition.X, newPosition.Y - oldPosition.Y);
                        }
                        break;
                        
                    case TouchAction.Up:
                        if (infoDictionary.ContainsKey(deviceId))
                            touchThumb = infoDictionary[deviceId].TouchThumb;

                        if (touchThumb != null)
                        {
                            Point oldPosition = infoDictionary[deviceId].Position;
                            Point newPosition = touchPoint.Position;

                            touchThumb.OnDragCompleted(newPosition.X - oldPosition.X, newPosition.Y - oldPosition.Y, false);
                            infoDictionary.Remove(touchPoint.TouchDevice.Id);
                        }
                        break;
                }
            }
        }

        static TouchThumb GetTouchThumbAncestor(DependencyObject child)
        {
            if (child == null)
                return null;

            if (child is TouchThumb)
                return child as TouchThumb;

            return GetTouchThumbAncestor(VisualTreeHelper.GetParent(child));
        }

        // Instance constructor adds instance to static collection
        public TouchThumb()
        {
            touchThumbs.Add(this);
        }

        // Mouse stuff is pretty standard
        protected override void OnMouseLeftButtonDown(MouseButtonEventArgs args)
        {
            mousePosition = args.GetPosition(this);
            OnDragStarted(mousePosition.X, mousePosition.Y);

            mousePosition = args.GetPosition(null);
            CaptureMouse();
            isMouseCaptured = true;
            args.Handled = true;

            base.OnMouseLeftButtonDown(args);
        }

        protected override void OnMouseMove(MouseEventArgs args)
        {
            if (isMouseCaptured)
            {
                Point newMousePosition = args.GetPosition(null);
                OnDragDelta(newMousePosition.X - mousePosition.X, newMousePosition.Y - mousePosition.Y);
                mousePosition = newMousePosition;
            }
            base.OnMouseMove(args);
        }

        protected override void OnMouseLeftButtonUp(MouseButtonEventArgs args)
        {
            if (isMouseCaptured)
            {
                isMouseCaptured = false;
                ReleaseMouseCapture();
            }

            base.OnMouseLeftButtonUp(args);
        }

        protected override void OnLostMouseCapture(MouseEventArgs args)
        {
            Point newMousePosition = args.GetPosition(null);
            OnDragCompleted(newMousePosition.X - mousePosition.X, newMousePosition.Y - mousePosition.Y, isMouseCaptured);
            isMouseCaptured = false;

            base.OnLostMouseCapture(args);
        }

        // Overridable methods fire events
        protected void OnDragStarted(double horizontalOffset, double verticalOffset)
        {
            if (DragStarted != null)
                DragStarted(this, new DragStartedEventArgs(horizontalOffset, verticalOffset));
        }

        protected void OnDragDelta(double horizontalChange, double verticalChange)
        {
            if (DragDelta != null)
                DragDelta(this, new DragDeltaEventArgs(horizontalChange, verticalChange));
        }

        protected void OnDragCompleted(double horizontalChange, double verticalChange, bool canceled)
        {
            if (DragCompleted != null)
                DragCompleted(this, new DragCompletedEventArgs(horizontalChange, verticalChange, canceled));
        }

        class Info
        {
            public TouchThumb TouchThumb { set; get; }
            public Point Position { set; get; }
        }
    }
}
