﻿// TouchAndDeformText (c) 2009 by Charles Petzold
using System;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Controls.Primitives;
using System.Windows.Input;
using System.Windows.Media;
using Petzold.Controls;

namespace TouchAndDeformText
{
    public partial class MainPage : UserControl
    {
        PathGeometry basePathGeometry;
        Rect boundingRect;
        bool isManipulating;
        int figureIndex = 0;

        public MainPage()
        {
            InitializeComponent();

            // Clone the PathGeometry and get its bounding box
            basePathGeometry = CloneFlattenedPathGeometry(textPathGeometry, ref boundingRect);

            // Initialize the thumb positions
            Canvas.SetLeft(thumbUpper0, boundingRect.Left);
            Canvas.SetLeft(thumbLower0, boundingRect.Left);
            Canvas.SetLeft(thumbUpper1, boundingRect.Left + boundingRect.Width / 3);
            Canvas.SetLeft(thumbLower1, boundingRect.Left + boundingRect.Width / 3);
            Canvas.SetLeft(thumbUpper2, boundingRect.Left + 2 * boundingRect.Width / 3);
            Canvas.SetLeft(thumbLower2, boundingRect.Left + 2 * boundingRect.Width / 3);
            Canvas.SetLeft(thumbUpper3, boundingRect.Right);
            Canvas.SetLeft(thumbLower3, boundingRect.Right);

            Canvas.SetTop(thumbUpper0, boundingRect.Top);
            Canvas.SetTop(thumbUpper1, boundingRect.Top);
            Canvas.SetTop(thumbUpper2, boundingRect.Top);
            Canvas.SetTop(thumbUpper3, boundingRect.Top);

            Canvas.SetTop(thumbLower0, boundingRect.Bottom);
            Canvas.SetTop(thumbLower1, boundingRect.Bottom);
            Canvas.SetTop(thumbLower2, boundingRect.Bottom);
            Canvas.SetTop(thumbLower3, boundingRect.Bottom);

            // Sync the Bezier curves with the thumbs
            SyncThumbsAndBeziers();

            TouchThumb.UnhandledFrameReported += OnTouchThumbUnhandledFrameReported;
        }

        // These two methods switch to manipulation mode and back
        void OnTouchThumbUnhandledFrameReported(object sender, TouchFrameEventArgs args)
        {
            ToggleManipulationMode(touchManipulateTemplate);
        }

        // Toggle between manipulating and not manipulating
        protected override void OnMouseLeftButtonDown(MouseButtonEventArgs args)
        {
            ToggleManipulationMode(mouseManipulateTemplate);
        }

        void ToggleManipulationMode(ControlTemplate manipulateTemplate)
        {
            if (!isManipulating)
            {
                // Assign the correct template
                foreach (UIElement element in manipulateCanvas.Children)
                    if (element is TouchThumb)
                        (element as TouchThumb).Template = manipulateTemplate;

                textPath.Opacity = 0.25;
                manipulateCanvas.Visibility = Visibility.Visible;
                isManipulating = true;
            }
            else
            {
                textPath.Opacity = 1;
                manipulateCanvas.Visibility = Visibility.Collapsed;
                isManipulating = false;
            }
        }

        // Move the thumbs and do everything a moved thumb implies
        void OnThumbDragDelta(object sender, DragDeltaEventArgs args)
        {
            ThumbDragCommon(sender as TouchThumb, args.HorizontalChange, args.VerticalChange);
            DeformText(figureIndex);
            figureIndex = (figureIndex + 1) % textPathGeometry.Figures.Count;
        }

        void OnThumbDragCompleted(object sender, DragCompletedEventArgs args)
        {
            ThumbDragCommon(sender as TouchThumb, args.HorizontalChange, args.VerticalChange);
            DeformText();
        }

        void ThumbDragCommon(TouchThumb thumb, double horizontalChange, double verticalChange)
        {
            Canvas.SetLeft(thumb, Canvas.GetLeft(thumb) + horizontalChange);
            Canvas.SetTop(thumb, Canvas.GetTop(thumb) + verticalChange);
            SyncThumbsAndBeziers();
        }

        // Set the Bezier points to the thumb locations
        void SyncThumbsAndBeziers()
        {
            upperBezierFigure.StartPoint = new Point(Canvas.GetLeft(thumbUpper0), Canvas.GetTop(thumbUpper0));
            upperBezierSegment.Point1    = new Point(Canvas.GetLeft(thumbUpper1), Canvas.GetTop(thumbUpper1));
            upperBezierSegment.Point2    = new Point(Canvas.GetLeft(thumbUpper2), Canvas.GetTop(thumbUpper2));
            upperBezierSegment.Point3    = new Point(Canvas.GetLeft(thumbUpper3), Canvas.GetTop(thumbUpper3));

            lowerBezierFigure.StartPoint = new Point(Canvas.GetLeft(thumbLower0), Canvas.GetTop(thumbLower0));
            lowerBezierSegment.Point1    = new Point(Canvas.GetLeft(thumbLower1), Canvas.GetTop(thumbLower1));
            lowerBezierSegment.Point2    = new Point(Canvas.GetLeft(thumbLower2), Canvas.GetTop(thumbLower2));
            lowerBezierSegment.Point3    = new Point(Canvas.GetLeft(thumbLower3), Canvas.GetTop(thumbLower3));

            // Tried to use ElementName binding in XAML for this stuff, but got the informative AG_E_PARSER_BAD_PROPERTY_VALUE error
            upperBezierLine1.StartPoint = upperBezierFigure.StartPoint;
            upperBezierLine1.EndPoint = upperBezierSegment.Point1;
            upperBezierLine2.StartPoint = upperBezierSegment.Point3;
            upperBezierLine2.EndPoint = upperBezierSegment.Point2;

            lowerBezierLine1.StartPoint = lowerBezierFigure.StartPoint;
            lowerBezierLine1.EndPoint = lowerBezierSegment.Point1;
            lowerBezierLine2.StartPoint = lowerBezierSegment.Point3;
            lowerBezierLine2.EndPoint = lowerBezierSegment.Point2;
        }

        // Set textPathGeometry from basePathGeometry but deformed based on Bezier curves
        void DeformText()
        {
            for (int figureIndex = 0; figureIndex < textPathGeometry.Figures.Count; figureIndex++)
                DeformText(figureIndex);
        }

        void DeformText(int figureIndex)
        {
            PathFigure textPathFigure = textPathGeometry.Figures[figureIndex];
            PathFigure basePathFigure = basePathGeometry.Figures[figureIndex];
            textPathFigure.StartPoint = DeformPoint(basePathFigure.StartPoint);

            for (int segmentIndex = 0; segmentIndex < textPathFigure.Segments.Count; segmentIndex++)
            {
                PolyLineSegment textPolyLineSegment = textPathFigure.Segments[segmentIndex] as PolyLineSegment;
                PolyLineSegment basePolyLineSegment = basePathFigure.Segments[segmentIndex] as PolyLineSegment;

                // Decouple PointCollection from textPolyLineSegment to improve performance
                PointCollection textPoints = textPolyLineSegment.Points;
                PointCollection basePoints = basePolyLineSegment.Points;
                textPolyLineSegment.Points = null;

                for (int pointIndex = 0; pointIndex < textPoints.Count; pointIndex++)
                {
                    textPoints[pointIndex] = DeformPoint(basePoints[pointIndex]);
                }

                // Recouple
                textPolyLineSegment.Points = textPoints;
            }
        }

        Point DeformPoint(Point basePoint)
        {
            // Try to cut down on the calculations by doing these first
            double t = (basePoint.X - boundingRect.Left) / boundingRect.Width;
            double mt = 1 - t;
            double t0 = mt * mt * mt;
            double t1 = 3 * t * mt * mt;
            double t2 = 3 * t * t * mt;
            double t3 = t * t * t;
            double w = (basePoint.Y - boundingRect.Top) / boundingRect.Height;
            double mw = 1 - w;

            double xUpper = t0 * upperBezierFigure.StartPoint.X +
                            t1 * upperBezierSegment.Point1.X +
                            t2 * upperBezierSegment.Point2.X +
                            t3 * upperBezierSegment.Point3.X;

            double yUpper = t0 * upperBezierFigure.StartPoint.Y +
                            t1 * upperBezierSegment.Point1.Y +
                            t2 * upperBezierSegment.Point2.Y +
                            t3 * upperBezierSegment.Point3.Y;

            double xLower = t0 * lowerBezierFigure.StartPoint.X +
                            t1 * lowerBezierSegment.Point1.X +
                            t2 * lowerBezierSegment.Point2.X +
                            t3 * lowerBezierSegment.Point3.X;

            double yLower = t0 * lowerBezierFigure.StartPoint.Y +
                            t1 * lowerBezierSegment.Point1.Y +
                            t2 * lowerBezierSegment.Point2.Y +
                            t3 * lowerBezierSegment.Point3.Y;

            return new Point(xUpper * mw + xLower * w,
                             yUpper * mw + yLower * w);
        }

        static PathGeometry CloneFlattenedPathGeometry(PathGeometry pathGeometry, ref Rect boundingRect)
        {
            double xMin = Double.MaxValue;
            double yMin = Double.MaxValue;
            double xMax = Double.MinValue;
            double yMax = Double.MinValue;

            PathGeometry clonedPathGeometry = new PathGeometry()
            {
                FillRule = pathGeometry.FillRule
            };

            foreach (PathFigure pathFigure in pathGeometry.Figures)
            {
                PathFigure clonedPathFigure = new PathFigure()
                {
                    IsClosed = pathFigure.IsClosed,
                    IsFilled = pathFigure.IsFilled,
                    StartPoint = pathFigure.StartPoint
                };

                NewMinMax(ref xMin, ref xMax, ref yMin, ref yMax, clonedPathFigure.StartPoint);

                foreach (PathSegment pathSegment in pathFigure.Segments)
                {
                    PolyLineSegment polyLineSegment = pathSegment as PolyLineSegment;

                    if (polyLineSegment == null)
                        throw new NotImplementedException("CloneFlattenedPathGeometry does not support PathSegment " +
                                                          "objects of type " + pathSegment.GetType().Name);

                    PolyLineSegment clonedPolyLineSegment = new PolyLineSegment();

                    foreach (Point point in polyLineSegment.Points)
                    {
                        NewMinMax(ref xMin, ref xMax, ref yMin, ref yMax, point);
                        clonedPolyLineSegment.Points.Add(point);
                    }
                    clonedPathFigure.Segments.Add(clonedPolyLineSegment);
                }
                clonedPathGeometry.Figures.Add(clonedPathFigure);
            }
            boundingRect = new Rect(new Point(xMin, yMin), new Point(xMax, yMax));
            return clonedPathGeometry;
        }

        static void NewMinMax(ref double xMin, ref double xMax, ref double yMin, ref double yMax, Point pt)
        {
            xMin = Math.Min(xMin, pt.X);
            xMax = Math.Max(xMax, pt.X);
            yMin = Math.Min(yMin, pt.Y);
            yMax = Math.Max(yMax, pt.Y);
        }

    }
}
